/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment media rendering library
 *
 * Copyright © 2006, 2007 Fluendo Embedded S.L.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author(s): Loïc Molinari <loic@fluendo.com>
 *            Julien Moutte <julien@fluendo.com>
 */

/**
 * SECTION:pgmdrawable
 * @short_description: A base class for all drawables.
 * @see_also: #PgmImage, #PgmText, #PgmCanvas.
 *
 * <refsect2>
 * <para>
 * #PgmDrawable represents a visual object on which you can draw on.
 * It is meant to be a base class for more specific visual objects such as
 * #PgmImage or #PgmText. Drawables are not visible until they are added to
 * a #PgmCanvas. A drawable never knows about physical on screen coordinates
 * such as pixels, and use floating point canvas coordinates. Drawables are
 * positioned in in 3D and then projected on a #PgmViewport.
 * </para>
 * <title>Drawing on drawables</title>
 * <para>
 * The base class implements simple drawing methods that are common to the
 * different subclasses. For example if you want to draw a left aligned green
 * text you would do something like that:
 * <example id="pigment-drawable-functions">
 * <title>Calling drawing functions on a drawable</title>
 * <programlisting language="c">
 * pgm_drawable_set_fg_color (text, 0, 255, 0, 255);
 * pgm_drawable_set_bg_color (text, 255, 0, 0, 255);
 * pgm_text_set_label (PGM_TEXT (text), "Hello world !");
 * </programlisting>
 * </example>
 * You are basically using a mix of drawable calls and text calls on the
 * same object at different level of the object hierarchy.
 * </para>
 * <title>Signal connections</title>
 * <para>
 * The <link linkend="PgmDrawable-changed">changed</link> signal is fired
 * whenever a property of a drawable is changed.
 * </para>
 * </refsect2>
 *
 * <emphasis remap="it">Last reviewed on 2007-06-01 (0.1.5)</emphasis>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <string.h>
#include "pgmdrawable.h"
#include "pgmcanvas.h"
#include "pgmmarshal.h"
#include "pgmenumtypes.h"

GST_DEBUG_CATEGORY (pgm_drawable_debug);
#define GST_CAT_DEFAULT pgm_drawable_debug

static GstObjectClass *parent_class = NULL;
guint pgm_drawable_signals[PGM_DRAWABLE_LAST_SIGNAL] = { 0 };

G_DEFINE_ABSTRACT_TYPE (PgmDrawable, pgm_drawable, GST_TYPE_OBJECT);

static void
pgm_drawable_dispose (GObject *object)
{
  GST_CALL_PARENT (G_OBJECT_CLASS, dispose, (object));
}

static void
pgm_drawable_class_init (PgmDrawableClass *klass)
{
  GObjectClass *gobject_class = (GObjectClass *) klass;

  parent_class = g_type_class_peek_parent (klass);

  /**
   * PgmDrawable::changed:
   * @drawable: the #PgmDrawable
   * @property: the #PgmDrawableProperty changed in @drawable
   *
   * Will be emitted after @property of @drawable was changed.
   */
  pgm_drawable_signals[PGM_DRAWABLE_CHANGED] =
    g_signal_new ("changed", G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET (PgmDrawableClass, changed),
                  NULL, NULL, pgm_marshal_VOID__ENUM, G_TYPE_NONE,
                  1, PGM_TYPE_DRAWABLE_PROPERTY);

  gobject_class->dispose = GST_DEBUG_FUNCPTR (pgm_drawable_dispose);

  GST_DEBUG_CATEGORY_INIT (pgm_drawable_debug, "pgm_drawable", 0,
                           "drawable object");
}

static void
pgm_drawable_init (PgmDrawable *drawable)
{
  drawable->layer = PGM_DRAWABLE_UNBINDED;
  /* Position */
  drawable->x = 0.0f;
  drawable->y = 0.0f;
  drawable->z = 0.0f;
  /* Size */
  drawable->width = 1.0f;
  drawable->height = 1.0f;
  /* Colors */
  drawable->fg_r = 255;
  drawable->fg_g = 255;
  drawable->fg_b = 255;
  drawable->fg_a = 255;
  drawable->bg_r = 255;
  drawable->bg_g = 255;
  drawable->bg_b = 255;
  drawable->bg_a = 255;
  /* Opacity */
  drawable->opacity = 255;
  /* Arc */
  drawable->arc_width = 0.0f;
  drawable->arc_height = 0.0f;
  /* Stroke */
  drawable->stroke_width = 0.0f;
  drawable->stroke_r = 0;
  drawable->stroke_g = 0;
  drawable->stroke_b = 0;
  drawable->stroke_a = 0;
  /* Shadow */
  drawable->shadow_radius = 0.0f;
  drawable->shadow_offset_x = 0.0f;
  drawable->shadow_offset_y = 0.0f;
  drawable->shadow_r = 0;
  drawable->shadow_g = 0;
  drawable->shadow_b = 0;
  drawable->shadow_a = 0;
}

/**
 * pgm_drawable_hide:
 * @drawable: a #PgmDrawable object.
 *
 * Makes @drawable invisible.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_hide (PgmDrawable *drawable)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  GST_OBJECT_FLAG_UNSET (drawable, PGM_DRAWABLE_FLAG_VISIBLE);
  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->hide)
    ret = klass->hide (drawable);

  g_signal_emit (G_OBJECT (drawable),
                 pgm_drawable_signals[PGM_DRAWABLE_CHANGED], 0,
                 PGM_DRAWABLE_VISIBILITY);

  return ret;
}

/**
 * pgm_drawable_show:
 * @drawable: a #PgmDrawable object.
 *
 * Makes @drawable visible.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_show (PgmDrawable *drawable)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  GST_OBJECT_FLAG_SET (drawable, PGM_DRAWABLE_FLAG_VISIBLE);
  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->show)
    ret = klass->show (drawable);

  g_signal_emit (G_OBJECT (drawable),
                 pgm_drawable_signals[PGM_DRAWABLE_CHANGED], 0,
                 PGM_DRAWABLE_VISIBILITY);

  return ret;
}

/**
 * pgm_drawable_is_visible:
 * @drawable: a #PgmDrawable object.
 * @visible: a pointer to a #gboolean where the visible state of the drawable
 * is going to be stored.
 *
 * Retrieves whether @drawable is visible.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_is_visible (PgmDrawable *drawable,
                         gboolean *visible)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (visible != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  *visible = GST_OBJECT_FLAG_IS_SET (drawable, PGM_DRAWABLE_FLAG_VISIBLE);
  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_size:
 * @drawable: a #PgmDrawable object.
 * @width: the width of the canvas.
 * @height: the height of the canvas.
 *
 * Defines @drawable size in canvas units to @width x @height.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_size (PgmDrawable *drawable,
                       gfloat width,
                       gfloat height)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  drawable->width = width;
  drawable->height = height;
  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_size)
    ret = klass->set_size (drawable, width, height);

  g_signal_emit (G_OBJECT (drawable),
                 pgm_drawable_signals[PGM_DRAWABLE_CHANGED], 0,
                 PGM_DRAWABLE_SIZE);

  return ret;
}

/**
 * pgm_drawable_get_size:
 * @drawable: a #PgmDrawable object.
 * @width: a pointer to a #gfloat where the width of the drawable is going to be
 * stored.
 * @height: a pointer to a #gfloat where the height of the drawable is going to
 * be stored.
 *
 * Gets @drawable size in canvas units.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_size (PgmDrawable *drawable,
                       gfloat *width,
                       gfloat *height)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (width != NULL, PGM_ERROR_X);
  g_return_val_if_fail (height != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *width = drawable->width;
  *height = drawable->height;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_position:
 * @drawable: a #PgmDrawable object.
 * @x: the position of @drawable on the x axis.
 * @y: the position of @drawable on the y axis.
 * @z: the position of @drawable on the z axis.
 *
 * Sets position of @drawable in canvas units.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_position (PgmDrawable *drawable,
                           gfloat x,
                           gfloat y,
                           gfloat z)
{
  PgmError ret = PGM_ERROR_OK;
  gboolean reordered = FALSE;
  PgmDrawableClass *klass;
  GstObject *canvas;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->x = x;
  drawable->y = y;
  if (drawable->z != z)
    {
      drawable->z = z;
      reordered = TRUE;
    }

  canvas = GST_OBJECT_PARENT (drawable);

  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_position)
    ret = klass->set_position (drawable, x, y, z);

  /* Reorder the drawable inside the canvas in the case that z has changed */
  if (reordered && canvas)
    _pgm_canvas_update_order (PGM_CANVAS (canvas), drawable);

  g_signal_emit (G_OBJECT (drawable),
                 pgm_drawable_signals[PGM_DRAWABLE_CHANGED], 0,
                 PGM_DRAWABLE_POSITION);

  return ret;
}

/**
 * pgm_drawable_get_position:
 * @drawable: a #PgmDrawable object.
 * @x: a pointer to a #gfloat where the x position of @drawable is going to
 * be stored.
 * @y: a pointer to a #gfloat where the y position of @drawable is going to
 * be stored.
 * @z: a pointer to a #gfloat where the z position of @drawable is going to
 * be stored.
 *
 * Gets @drawable position in canvas units.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_position (PgmDrawable *drawable,
                           gfloat *x,
                           gfloat *y,
                           gfloat *z)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (x != NULL, PGM_ERROR_X);
  g_return_val_if_fail (y != NULL, PGM_ERROR_X);
  g_return_val_if_fail (z != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *x = drawable->x;
  *y = drawable->y;
  *z = drawable->z;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_fg_color:
 * @drawable: a #PgmDrawable object.
 * @r: the red foreground color.
 * @g: the green foreground color.
 * @b: the blue foreground color.
 * @a: the alpha foreground color.
 *
 * Set the color used for drawing operations on @drawable to (@r,@g,@b,@a).
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_fg_color (PgmDrawable *drawable,
                           guchar r,
                           guchar g,
                           guchar b,
                           guchar a)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->fg_r = r;
  drawable->fg_g = g;
  drawable->fg_b = b;
  drawable->fg_a = a;

  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_fg_color)
    ret = klass->set_fg_color (drawable, r, g, b, a);

  g_signal_emit (G_OBJECT (drawable),
                 pgm_drawable_signals[PGM_DRAWABLE_CHANGED], 0,
                 PGM_DRAWABLE_FG_COLOR);

  return ret;
}

/**
 * pgm_drawable_get_fg_color:
 * @drawable: a #PgmDrawable object.
 * @r: a pointer to a #guchar where the red foreground color is going to be
 * stored.
 * @g: a pointer to a #guchar where the green foreground color is going to be
 * stored.
 * @b: a pointer to a #guchar where the blue foreground color is going to be
 * stored.
 * @a: a pointer to a #guchar where the alpha foreground color is going to be
 * stored.
 *
 * Retrieves the color used for drawing operations on @drawable in
 * (@r,@g,@b,@a).
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_fg_color (PgmDrawable *drawable,
                           guchar *r,
                           guchar *g,
                           guchar *b,
                           guchar *a)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (r != NULL, PGM_ERROR_X);
  g_return_val_if_fail (g != NULL, PGM_ERROR_X);
  g_return_val_if_fail (b != NULL, PGM_ERROR_X);
  g_return_val_if_fail (a != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *r = drawable->fg_r;
  *g = drawable->fg_g;
  *b = drawable->fg_b;
  *a = drawable->fg_a;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_bg_color:
 * @drawable: a #PgmDrawable object.
 * @r: the red background color.
 * @g: the green background color.
 * @b: the blue background color.
 * @a: the alpha background color.
 *
 * Sets the color used to fill the background of @drawable to (@r,@g,@b,@a).
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_bg_color (PgmDrawable *drawable,
                           guchar r,
                           guchar g,
                           guchar b,
                           guchar a)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->bg_r = r;
  drawable->bg_g = g;
  drawable->bg_b = b;
  drawable->bg_a = a;

  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_bg_color)
    ret = klass->set_bg_color (drawable, r, g, b, a);

  g_signal_emit (G_OBJECT (drawable),
                 pgm_drawable_signals[PGM_DRAWABLE_CHANGED], 0,
                 PGM_DRAWABLE_BG_COLOR);

  return ret;
}

/**
 * pgm_drawable_get_bg_color:
 * @drawable: A #PgmDrawable object.
 * @r: a pointer to a #guchar where the red background color is going to be
 * stored.
 * @g: a pointer to a #guchar where the green background color is going to be
 * stored.
 * @b: a pointer to a #guchar where the blue background color is going to be
 * stored.
 * @a: a pointer to a #guchar where the alpha background color is going to be
 * stored.
 *
 * Retrieves the color used to fill the background of @drawable in
 * (@r,@g,@b,@a).
 *
 * MT safe.
 *
 * Returns: A #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_bg_color (PgmDrawable *drawable,
                           guchar *r,
                           guchar *g,
                           guchar *b,
                           guchar *a)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (r != NULL, PGM_ERROR_X);
  g_return_val_if_fail (g != NULL, PGM_ERROR_X);
  g_return_val_if_fail (b != NULL, PGM_ERROR_X);
  g_return_val_if_fail (a != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *r = drawable->bg_r;
  *g = drawable->bg_g;
  *b = drawable->bg_b;
  *a = drawable->bg_a;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_opacity:
 * @drawable: a #PgmDrawable object.
 * @opacity: the opacity of @drawable.
 *
 * Defines the opacity of @drawable to @opacity.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_opacity (PgmDrawable *drawable,
                          guchar opacity)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->opacity = opacity;

  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_opacity)
    ret = klass->set_opacity (drawable, opacity);

  g_signal_emit (G_OBJECT (drawable),
                 pgm_drawable_signals[PGM_DRAWABLE_CHANGED], 0,
                 PGM_DRAWABLE_OPACITY);

  return ret;
}

/**
 * pgm_drawable_get_opacity:
 * @drawable: a #PgmDrawable object.
 * @opacity: a pointer to a #guchar where the opacity is going to be stored.
 *
 * Retrieves the opacity of @drawable in @opacity.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_opacity (PgmDrawable *drawable,
                          guchar *opacity)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (opacity != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  *opacity = drawable->opacity;
  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_arc:
 * @drawable: a #PgmDrawable object.
 * @width: the width of the arc.
 * @height: the height of the arc.
 *
 * Sets the arc @width and @height appearing on the corners of @drawable. The
 * contents of @drawable will then be clipped on the corners.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_arc (PgmDrawable *drawable,
                      gfloat width,
                      gfloat height)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->arc_width = width;
  drawable->arc_height = height;

  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_arc)
    ret = klass->set_arc (drawable, width, height);

  g_signal_emit (G_OBJECT (drawable),
                 pgm_drawable_signals[PGM_DRAWABLE_CHANGED], 0,
                 PGM_DRAWABLE_ARC);

  return ret;
}

/**
 * pgm_drawable_get_arc:
 * @drawable: a #PgmDrawable object.
 * @width: a pointer to #gfloat where the width of the arc is going to be
 * stored.
 * @height: a pointer to #gfloat where the height of the arc is going to be
 * stored.
 *
 * Retrieves the arc width and height of @drawable in @width and @height.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_arc (PgmDrawable *drawable,
                      gfloat *width,
                      gfloat *height)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (width != NULL, PGM_ERROR_X);
  g_return_val_if_fail (height != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *width = drawable->arc_width;
  *height = drawable->arc_height;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/* Stroke methods */

/**
 * pgm_drawable_set_stroke_width:
 * @drawable: a #PgmDrawable object.
 * @width: the width of the stroke.
 *
 * Defines the @width of the stroke drawn around @drawable.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_stroke_width (PgmDrawable *drawable,
                               gfloat width)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->stroke_width = width;

  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_stroke_width)
    ret = klass->set_stroke_width (drawable, width);

  g_signal_emit (G_OBJECT (drawable),
                 pgm_drawable_signals[PGM_DRAWABLE_CHANGED], 0,
                 PGM_DRAWABLE_STROKE_WIDTH);

  return ret;
}

/**
 * pgm_drawable_get_stroke_width:
 * @drawable: a #PgmDrawable object.
 * @width: a pointer to #gfloat where the width of the stroke is going to be
 * stored.
 *
 * Retrieves the @width of the stroke drawn around @drawable.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_stroke_width (PgmDrawable *drawable,
                               gfloat *width)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (width != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *width = drawable->stroke_width;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_stroke_color:
 * @drawable: a #PgmDrawable object.
 * @r: the red stroke color.
 * @g: the green stroke color.
 * @b: the blue stroke color.
 * @a: the alpha stroke color.
 *
 * Sets the color of the stroke drawn around @drawable.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_stroke_color (PgmDrawable *drawable,
                               guchar r,
                               guchar g,
                               guchar b,
                               guchar a)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass * klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->stroke_r = r;
  drawable->stroke_g = g;
  drawable->stroke_b = b;
  drawable->stroke_a = a;

  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_stroke_color)
    ret = klass->set_stroke_color (drawable, r, g, b, a);

  g_signal_emit (G_OBJECT (drawable),
                 pgm_drawable_signals[PGM_DRAWABLE_CHANGED], 0,
                 PGM_DRAWABLE_STROKE_COLOR);

  return ret;
}

/**
 * pgm_drawable_get_stroke_color:
 * @drawable: a #PgmDrawable object.
 * @r: a pointer to #guchar where the red stroke color is going to be stored.
 * @g: a pointer to #guchar where the green stroke color is going to be stored.
 * @b: a pointer to #guchar where the blue stroke color is going to be stored.
 * @a: a pointer to #guchar where the alpha stroke color is going to be stored.
 *
 * Sets the color of the stroke drawn around @drawable.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_stroke_color (PgmDrawable *drawable,
                               guchar *r,
                               guchar *g,
                               guchar *b,
                               guchar *a)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (r != NULL, PGM_ERROR_X);
  g_return_val_if_fail (g != NULL, PGM_ERROR_X);
  g_return_val_if_fail (b != NULL, PGM_ERROR_X);
  g_return_val_if_fail (a != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *r = drawable->stroke_r;
  *g = drawable->stroke_g;
  *b = drawable->stroke_b;
  *a = drawable->stroke_a;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/* Shadow methods */

/**
 * pgm_drawable_set_shadow_radius:
 * @drawable: a #PgmDrawable object.
 * @radius: the radius of the shadow.
 *
 * Defines the @radius of the shadow drawn under @drawable. The radius defines
 * the width of the shadow drawn on each sides of the @drawable.
 *
 * MT safe.
 *
 * Returns: A #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_shadow_radius (PgmDrawable *drawable,
                                gfloat radius)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->shadow_radius = radius;

  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_shadow_radius)
    ret = klass->set_shadow_radius (drawable, radius);

  g_signal_emit (G_OBJECT (drawable),
                 pgm_drawable_signals[PGM_DRAWABLE_CHANGED], 0,
                 PGM_DRAWABLE_SHADOW_RADIUS);

  return ret;
}

/**
 * pgm_drawable_get_shadow_radius:
 * @drawable: a #PgmDrawable object.
 * @radius: a pointer to #gfloat where the radius of the shadow is going to be
 * stored.
 *
 * Retrieves the @radius of the shadow drawn under @drawable.
 *
 * MT safe.
 *
 * Returns: A #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_shadow_radius (PgmDrawable *drawable,
                                gfloat *radius)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (radius != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *radius = drawable->shadow_radius;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_shadow_offset:
 * @drawable: a #PgmDrawable object.
 * @x: the x offset. 0.0 by default.
 * @y: the y offset. 0.0 by default.
 *
 * Defines the (@x,@y) offset in canvas coordinates applied to the shadow
 * relative to the @drawable.
 *
 * MT safe.
 *
 * Returns: A #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_shadow_offset (PgmDrawable *drawable,
                                gfloat x,
                                gfloat y)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->shadow_offset_x = x;
  drawable->shadow_offset_y = y;

  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_shadow_offset)
    ret = klass->set_shadow_offset (drawable, x, y);

  g_signal_emit (G_OBJECT (drawable),
                 pgm_drawable_signals[PGM_DRAWABLE_CHANGED], 0,
                 PGM_DRAWABLE_SHADOW_OFFSET);

  return ret;
}

/**
 * pgm_drawable_get_shadow_offset:
 * @drawable: a #PgmDrawable object.
 * @x: a pointer to #gfloat where the x offset is going to be stored.
 * @y: a pointer to #gfloat where the y offset is going to be stored.
 *
 * Retrieves the (@x,@y) offset applied to the shadow relative to the
 * @drawable.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_shadow_offset (PgmDrawable *drawable,
                                gfloat *x,
                                gfloat *y)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (x != NULL, PGM_ERROR_X);
  g_return_val_if_fail (y != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *x = drawable->shadow_offset_x;
  *y = drawable->shadow_offset_y;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_shadow_color:
 * @drawable: a #PgmDrawable object.
 * @r: the red shadow color component.
 * @g: the green shadow color component.
 * @b: the blue shadow color component.
 * @a: the alpha shadow color component.
 *
 * Sets the color of the shadow drawn under @drawable.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_shadow_color (PgmDrawable *drawable,
                               guchar r,
                               guchar g,
                               guchar b,
                               guchar a)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->shadow_r = r;
  drawable->shadow_g = g;
  drawable->shadow_b = b;
  drawable->shadow_a = a;

  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_shadow_color)
    ret = klass->set_shadow_color (drawable, r, g, b, a);

  g_signal_emit (G_OBJECT (drawable),
                 pgm_drawable_signals[PGM_DRAWABLE_CHANGED], 0,
                 PGM_DRAWABLE_SHADOW_COLOR);

  return ret;
}

/**
 * pgm_drawable_get_shadow_color:
 * @drawable: a #PgmDrawable object.
 * @r: a pointer to #guchar where the red shadow color component  is going to
 * be stored.
 * @g: a pointer to #guchar where the green shadow color component is going to
 * be stored.
 * @b: a pointer to #guchar where the blue shadow color component is going to
 * be stored.
 * @a: a pointer to #guchar where the alpha shadow color component is going to
 * be stored.
 *
 * Retrieves the color of the shadow drawn under @drawable.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_shadow_color (PgmDrawable *drawable,
                               guchar *r,
                               guchar *g,
                               guchar *b,
                               guchar *a)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (r != NULL, PGM_ERROR_X);
  g_return_val_if_fail (g != NULL, PGM_ERROR_X);
  g_return_val_if_fail (b != NULL, PGM_ERROR_X);
  g_return_val_if_fail (a != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *r = drawable->shadow_r;
  *g = drawable->shadow_g;
  *b = drawable->shadow_b;
  *a = drawable->shadow_a;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}
