/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment OpenGL plugin
 *
 * Copyright © 2006, 2007 Fluendo Embedded S.L.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author: Loïc Molinari <loic@fluendo.com>
 */

/*
 * The plugin duplicates all the generic Pigment drawables in GL drawables.
 * The GL drawables are stored in layer lists inside this GL viewport. The
 * generic layer lists are duplicated when the user calls the set_canvas
 * method. The duplicated GL drawables are deleted when the user deletes the
 * canvas or binds a NULL canvas to the viewport. Thanks to Canvas 'added' and
 * 'removed' signals, the GL drawable lists can also reflect correctly the
 * generic ones.
 *
 * When a generic drawable property changes, it emits a signal so that the
 * GL drawable can be easily adpated accordingly. The link between the generic
 * and the GL drawable is done through a dedicated hash-table using the generic
 * drawable address as the key and the GL drawable address as the value.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include "pgmglviewport.h"
#include "pgmglimage.h"
#include "pgmgltext.h"

GST_DEBUG_CATEGORY_EXTERN (pgm_gl_debug);
#define GST_CAT_DEFAULT pgm_gl_debug

static PgmViewportClass *parent_class = NULL;

/* Private functions */

/* Drawable 'changed' handler */
static void
changed_cb (PgmDrawable *drawable,
            PgmDrawableProperty property,
            gpointer data)
{
  PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (data);
  PgmGlViewportClass *klass;

  if (gldrawable->glviewport)
    {
      klass = PGM_GL_VIEWPORT_GET_CLASS (gldrawable->glviewport);
      klass->changed_func[property] (gldrawable);
    }
}

/* Create a GlDrawable from a generic Drawable */
static PgmGlDrawable *
create_gl_drawable (PgmGlViewport *glviewport,
                    PgmDrawable *drawable)
{
  PgmGlDrawable *gldrawable = NULL;

  /* Create the GlDrawable depending on the type */
  if (PGM_IS_IMAGE (drawable))
    {
      gldrawable = pgm_gl_image_new (drawable, glviewport);
      GST_DEBUG_OBJECT (glviewport, "created %s", GST_OBJECT_NAME (gldrawable));
    }
  else if (PGM_IS_TEXT (drawable))
    {
      gldrawable = pgm_gl_text_new (drawable, glviewport);
      GST_DEBUG_OBJECT (glviewport, "created %s", GST_OBJECT_NAME (gldrawable));
    }
  else
    {
      gldrawable = NULL;
      GST_DEBUG_OBJECT (glviewport, "cannot create object from this type");
    }

  /* Make it easily retrievable and connect to the 'changed' signal so that
   * changes could be reflected in the plugin */
  if (gldrawable)
    {
      GST_OBJECT_LOCK (glviewport);
      GST_OBJECT_LOCK (drawable);

      g_hash_table_insert (glviewport->drawable_hash, drawable, gldrawable);
      gldrawable->change_handler = g_signal_connect (drawable, "changed",
                                                     G_CALLBACK (changed_cb),
                                                     (gpointer) gldrawable);

      GST_OBJECT_UNLOCK (drawable);
      GST_OBJECT_UNLOCK (glviewport);
    }

  return gldrawable;
}

/* Delete a GlDrawable */
static void
delete_gl_drawable (PgmGlViewport *glviewport,
                    PgmGlDrawable *gldrawable)
{
  /* Disconnect from the 'changed' signal and remove it from the hash */
  GST_OBJECT_LOCK (glviewport);
  GST_OBJECT_LOCK (gldrawable->drawable);

  g_signal_handler_disconnect (gldrawable->drawable, gldrawable->change_handler);
  g_hash_table_remove (glviewport->drawable_hash, gldrawable->drawable);

  GST_OBJECT_UNLOCK (gldrawable->drawable);
  GST_OBJECT_UNLOCK (glviewport);

  /* Then delete it */
  GST_DEBUG_OBJECT (glviewport, "deleting %s", GST_OBJECT_NAME (gldrawable));
  gst_object_unref (gldrawable);
}

/* GCompareFunc used to compare z-position at insertion in the layer */
static gint
z_compare (PgmGlDrawable *gldrawable1,
           PgmGlDrawable *gldrawable2)
{
  if (gldrawable1->drawable->z < gldrawable2->drawable->z)
    return -1;
  else
    return 1;
}

/* Canvas 'drawable-added' handler */
static void
drawable_added_cb (PgmCanvas *canvas,
                   PgmDrawable *drawable,
                   PgmDrawableLayer layer,
                   gint order,
                   gpointer data)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (data);
  PgmGlDrawable *gldrawable;

  GST_DEBUG_OBJECT (glviewport, "drawable_added_cb");

  /* Create the GlDrawable */
  gldrawable = create_gl_drawable (glviewport, drawable);

  /* And insert it in the correct layer */
  GST_OBJECT_LOCK (glviewport);
  switch (layer)
    {
    case PGM_DRAWABLE_NEAR:
      glviewport->near = g_list_insert_sorted (glviewport->near, gldrawable,
                                               (GCompareFunc) z_compare);
      break;

    case PGM_DRAWABLE_MIDDLE:
      glviewport->middle = g_list_insert_sorted (glviewport->middle, gldrawable,
                                                 (GCompareFunc) z_compare);
      break;

    case PGM_DRAWABLE_FAR:
      glviewport->far = g_list_insert_sorted (glviewport->far, gldrawable,
                                              (GCompareFunc) z_compare);
      break;

    default:
      break;
    }
  GST_OBJECT_UNLOCK (glviewport);
}

/* Canvas 'drawable-removed' handler */
static void
drawable_removed_cb (PgmCanvas *canvas,
                     PgmDrawable *drawable,
                     PgmDrawableLayer layer,
                     gpointer data)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (data);
  PgmGlDrawable *gldrawable;

  GST_DEBUG_OBJECT (glviewport, "drawable_removed_cb");

  GST_OBJECT_LOCK (glviewport);

  /* Get the GlDrawable */
  gldrawable = g_hash_table_lookup (glviewport->drawable_hash, drawable);

  /* Remove it from the list */
  switch (layer)
    {
    case PGM_DRAWABLE_NEAR:
      glviewport->near = g_list_remove (glviewport->near, gldrawable);
      break;

    case PGM_DRAWABLE_MIDDLE:
      glviewport->middle = g_list_remove (glviewport->middle, gldrawable);
      break;

    case PGM_DRAWABLE_FAR:
      glviewport->far = g_list_remove (glviewport->far, gldrawable);
      break;

    default:
      break;
    }

  GST_OBJECT_UNLOCK (glviewport);

  /* And delete it */
  delete_gl_drawable (glviewport, gldrawable);
}

/* Canvas 'drawable-reordered' handler */
static void
drawable_reordered_cb (PgmCanvas *canvas,
                       PgmDrawable *drawable,
                       PgmDrawableLayer layer,
                       gint order,
                       gpointer data)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (data);
  PgmGlDrawable *gldrawable;

  GST_DEBUG_OBJECT (glviewport, "drawable_reordered_cb");

  GST_OBJECT_LOCK (glviewport);

  /* Get the GlDrawable */
  gldrawable = g_hash_table_lookup (glviewport->drawable_hash, drawable);

  /* And reorder it */
  switch (layer)
    {
    case PGM_DRAWABLE_NEAR:
      glviewport->near = g_list_remove (glviewport->near, gldrawable);
      glviewport->near = g_list_insert (glviewport->near, gldrawable, order);
      break;

    case PGM_DRAWABLE_MIDDLE:
      glviewport->middle = g_list_remove (glviewport->middle, gldrawable);
      glviewport->middle = g_list_insert (glviewport->middle, gldrawable, order);
      break;

    case PGM_DRAWABLE_FAR:
      glviewport->far = g_list_remove (glviewport->far, gldrawable);
      glviewport->far = g_list_insert (glviewport->far, gldrawable, order);
      break;

    default:
      break;
    }

  GST_OBJECT_UNLOCK (glviewport);
}

/* Canvas 'size-changed' handler */
static void
size_changed_cb (PgmCanvas *canvas,
                 gpointer data)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (data);
  PgmContextTask *task;
  gfloat non_square_width;

  GST_OBJECT_LOCK (canvas);
  GST_OBJECT_LOCK (glviewport);

  non_square_width = canvas->width * glviewport->non_square_ratio;
  glviewport->canvas_ratio = non_square_width / canvas->height;

  GST_OBJECT_UNLOCK (glviewport);
  GST_OBJECT_UNLOCK (canvas);

  task = pgm_context_task_new (PGM_CONTEXT_PROJECTION, NULL);
  pgm_context_push_deferred_task (glviewport->context, task);
}

/* Canvas 'regenerated' handler */
static void
regenerated_cb (PgmCanvas *canvas,
                gpointer data)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (data);
  GList *list = NULL;

  list = glviewport->near;
  while (list)
    {
      pgm_gl_drawable_regenerate (PGM_GL_DRAWABLE (list->data));
      list = list->next;
    }

  list = glviewport->middle;
  while (list)
    {
      pgm_gl_drawable_regenerate (PGM_GL_DRAWABLE (list->data));
      list = list->next;
    }

  list = glviewport->far;
  while (list)
    {
      pgm_gl_drawable_regenerate (PGM_GL_DRAWABLE (list->data));
      list = list->next;
    }
}

/* Delete all GlDrawable from the given GlViewport layer list and free it */
static void
delete_all_gl_drawable_from_layer (PgmGlViewport *glviewport,
                                   GList **layer_list)
{
  GList *list = *layer_list;

  while (list)
    {
      delete_gl_drawable (glviewport, (PgmGlDrawable *) list->data);
      list = list->next;
    }

  GST_OBJECT_LOCK (glviewport);
  g_list_free (*layer_list);
  *layer_list = NULL;
  GST_OBJECT_UNLOCK (glviewport);
}

/* Synchronize the Drawables from the given canvas creating and adding the
 * the corresponding GlDrawables in the layers */
static void
sync_gl_drawable_from_canvas (PgmGlViewport *glviewport,
                              PgmCanvas *canvas)
{
  PgmGlDrawable *gldrawable;
  GList *list;

  /* Sync near layer */
  list = canvas->near;
  while (list)
    {
      gldrawable = create_gl_drawable (glviewport, (PgmDrawable *) list->data);
      GST_OBJECT_LOCK (canvas);
      GST_OBJECT_LOCK (glviewport);
      glviewport->near = g_list_append (glviewport->near, gldrawable);
      GST_OBJECT_UNLOCK (glviewport);
      GST_OBJECT_UNLOCK (canvas);
      list = list->next;
    }

  /* Sync middle layer */
  list = canvas->middle;
  while (list)
    {
      gldrawable = create_gl_drawable (glviewport, (PgmDrawable *) list->data);
      GST_OBJECT_LOCK (canvas);
      GST_OBJECT_LOCK (glviewport);
      glviewport->middle = g_list_append (glviewport->middle, gldrawable);
      GST_OBJECT_UNLOCK (glviewport);
      GST_OBJECT_UNLOCK (canvas);
      list = list->next;
    }

  /* Sync far layer */
  list = canvas->far;
  while (list)
    {
      gldrawable = create_gl_drawable (glviewport, (PgmDrawable *) list->data);
      GST_OBJECT_LOCK (canvas);
      GST_OBJECT_LOCK (glviewport);
      glviewport->far = g_list_append (glviewport->far, gldrawable);
      GST_OBJECT_UNLOCK (glviewport);
      GST_OBJECT_UNLOCK (canvas);
      list = list->next;
    }
}

/* Defines the non-square handling parameters given a screen size */
static void
update_non_square_parameters (PgmGlViewport *glviewport,
                              guint screen_width,
                              guint screen_height)
{
  PgmCanvas *canvas = PGM_VIEWPORT (glviewport)->canvas;
  PgmContextTask *task;
  gint resolution_width, resolution_height;

  pgm_backend_get_screen_resolution (glviewport->context->backend,
                                     &resolution_width, &resolution_height);

  GST_OBJECT_LOCK (glviewport);

  glviewport->non_square_ratio = ((gfloat) resolution_width / resolution_height)
    * ((gfloat) screen_height / screen_width);

  /* Adapt the canvas_ratio */
  if (canvas)
    glviewport->canvas_ratio = (canvas->width / canvas->height)
      * glviewport->non_square_ratio;
  else
    glviewport->canvas_ratio = 1.333333334f * glviewport->non_square_ratio;

  GST_OBJECT_UNLOCK (glviewport);

  /* And push a projection update */
  task = pgm_context_task_new (PGM_CONTEXT_PROJECTION, NULL);
  pgm_context_push_immediate_task (glviewport->context, task);
}

/* PgmViewport methods */

static PgmError
pgm_gl_viewport_set_color (PgmViewport *viewport,
                           guchar r,
                           guchar g,
                           guchar b,
                           guchar a)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
  PgmContextTask *task;

  GST_DEBUG_OBJECT (glviewport, "set_color");

  task = pgm_context_task_new (PGM_CONTEXT_CLEAR_COLOR, NULL);
  pgm_context_push_deferred_task (glviewport->context, task);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_update (PgmViewport *viewport)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);

  GST_DEBUG_OBJECT (glviewport, "update");

  pgm_context_update (glviewport->context);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_set_title (PgmViewport *viewport,
                           const gchar * title)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
  PgmContextTask *task;

  GST_DEBUG_OBJECT (glviewport, "set_title");

  task = pgm_context_task_new (PGM_CONTEXT_TITLE, NULL);
  pgm_context_push_immediate_task (glviewport->context, task);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_show (PgmViewport *viewport)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);

  GST_DEBUG_OBJECT (glviewport, "show");

  pgm_backend_show (glviewport->context->backend);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_hide (PgmViewport *viewport)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);

  GST_DEBUG_OBJECT (glviewport, "hide");

  pgm_backend_hide (glviewport->context->backend);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_set_cursor (PgmViewport *viewport,
                            PgmViewportCursor cursor)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
  PgmContextTask *task;

  GST_DEBUG_OBJECT (glviewport, "set_cursor");

  task = pgm_context_task_new (PGM_CONTEXT_CURSOR, NULL);
  pgm_context_push_immediate_task (glviewport->context, task);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_set_icon (PgmViewport *viewport,
                          GdkPixbuf *icon)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
  PgmContextTask *task;

  GST_DEBUG_OBJECT (glviewport, "set_icon");

  task = pgm_context_task_new (PGM_CONTEXT_ICON, NULL);
  pgm_context_push_immediate_task (glviewport->context, task);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_set_size (PgmViewport *viewport,
                          gint width,
                          gint height)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
  PgmContextTask *task;

  GST_DEBUG_OBJECT (glviewport, "set_size");

  GST_OBJECT_LOCK (glviewport);
  glviewport->viewport_ratio = (gfloat) viewport->width / viewport->height;
  GST_OBJECT_UNLOCK (glviewport);

  task = pgm_context_task_new (PGM_CONTEXT_SIZE, NULL);
  pgm_context_push_immediate_task (glviewport->context, task);

  task = pgm_context_task_new (PGM_CONTEXT_PROJECTION, NULL);
  pgm_context_push_deferred_task (glviewport->context, task);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_set_fullscreen (PgmViewport *viewport,
                                gboolean fullscreen)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
  PgmContextTask *task;

  GST_DEBUG_OBJECT (glviewport, "set_fullscreen");

  task = pgm_context_task_new (PGM_CONTEXT_FULLSCREEN, NULL);
  pgm_context_push_immediate_task (glviewport->context, task);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_set_screen_resolution (PgmViewport *viewport,
                                       gint width,
                                       gint height)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
  PgmContextTask *task;

  GST_DEBUG_OBJECT (glviewport, "set_screen_resolution");

  task = pgm_context_task_new (PGM_CONTEXT_RESOLUTION, NULL);
  pgm_context_push_immediate_task (glviewport->context, task);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_get_screen_resolution (PgmViewport *viewport,
                                       gint *width,
                                       gint *height)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);

  GST_DEBUG_OBJECT (glviewport, "get_screen_resolution");

  pgm_backend_get_screen_resolution (glviewport->context->backend, width,
                                     height);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_set_screen_size_mm (PgmViewport *viewport,
                                    gint width,
                                    gint height)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);

  GST_DEBUG_OBJECT (glviewport, "set_screen_size_mm");

  update_non_square_parameters (glviewport, width, height);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_get_screen_size_mm (PgmViewport *viewport,
                                    gint *width,
                                    gint *height)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);

  GST_DEBUG_OBJECT (glviewport, "get_screen_size_mm");

  pgm_backend_get_screen_size_mm (glviewport->context->backend, width, height);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_set_canvas (PgmViewport *viewport,
                            PgmCanvas *canvas)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
  gfloat non_square_width;

  GST_DEBUG_OBJECT (glviewport, "set_canvas");

  /* There's a canvas bound, let's clear associated data */
  if (glviewport->canvas)
    {
      GST_OBJECT_LOCK (glviewport);
      GST_OBJECT_LOCK (glviewport->canvas);

      /* Disconnect handlers */
      g_signal_handler_disconnect (glviewport->canvas,
                                   glviewport->add_handler);
      g_signal_handler_disconnect (glviewport->canvas,
                                   glviewport->remove_handler);
      g_signal_handler_disconnect (glviewport->canvas,
                                   glviewport->reorder_handler);
      g_signal_handler_disconnect (glviewport->canvas,
                                   glviewport->size_handler);
      g_signal_handler_disconnect (glviewport->canvas,
                                   glviewport->regenerated_handler);

      GST_OBJECT_UNLOCK (glviewport->canvas);
      GST_OBJECT_UNLOCK (glviewport);

      /* Delete our hierarchy */
      delete_all_gl_drawable_from_layer (glviewport, &glviewport->near);
      delete_all_gl_drawable_from_layer (glviewport, &glviewport->middle);
      delete_all_gl_drawable_from_layer (glviewport, &glviewport->far);

      GST_OBJECT_LOCK (glviewport);
      glviewport->canvas = NULL;
      GST_OBJECT_UNLOCK (glviewport);
    }

  /* Bind the new canvas if any */
  if (canvas)
    {
      sync_gl_drawable_from_canvas (glviewport, canvas);

      GST_OBJECT_LOCK (glviewport);
      GST_OBJECT_LOCK (canvas);

      /* Connect handlers */
      glviewport->add_handler =
        g_signal_connect (G_OBJECT (canvas), "drawable-added",
                          G_CALLBACK (drawable_added_cb),
                          (gpointer) glviewport);
      glviewport->remove_handler =
        g_signal_connect (G_OBJECT (canvas), "drawable-removed",
                          G_CALLBACK (drawable_removed_cb),
                          (gpointer) glviewport);
      glviewport->reorder_handler =
        g_signal_connect (G_OBJECT (canvas), "drawable-reordered",
                          G_CALLBACK (drawable_reordered_cb),
                          (gpointer) glviewport);
      glviewport->size_handler =
        g_signal_connect (G_OBJECT (canvas), "size-changed",
                          G_CALLBACK (size_changed_cb),
                          (gpointer) glviewport);
      glviewport->regenerated_handler =
        g_signal_connect (G_OBJECT (canvas), "regenerated",
                          G_CALLBACK (regenerated_cb),
                          (gpointer) glviewport);

      glviewport->canvas = canvas;
      non_square_width = canvas->width * glviewport->non_square_ratio;
      glviewport->canvas_ratio = non_square_width / canvas->height;

      GST_OBJECT_UNLOCK (canvas);
      GST_OBJECT_UNLOCK (glviewport);
    }

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_get_pixel_formats (PgmViewport *viewport,
                                   gulong *formats_mask)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
  gchar *variable;

  GST_DEBUG_OBJECT (glviewport, "get_pixel_formats");

  GST_OBJECT_LOCK (glviewport);

  /* The classic formats supported by all hardware */
  *formats_mask = PGM_IMAGE_RGB
                | PGM_IMAGE_RGBA
                | PGM_IMAGE_BGR
                | PGM_IMAGE_BGRA;

  /* I420 and YV12 are supported through fragment programs and
   * multi-texturing with a minimum texture units number of three. The
   * PGM_GL_CSP_PROGRAM environment variable allows to disable the use
   * fragment programs to do the color space conversion. */
  if (glviewport->context->feature_mask & PGM_GL_FEAT_PER_PLANE_YCBCR_PROGRAM)
    {
      *formats_mask |= PGM_IMAGE_I420 | PGM_IMAGE_YV12;

      variable = getenv ("PGM_GL_CSP_PROGRAM");
      if (variable && variable[0] == '0')
        *formats_mask &= ~(PGM_IMAGE_I420 | PGM_IMAGE_YV12);
    }

  /* FIXME: YUYV and UYVY are not supported */

  GST_OBJECT_UNLOCK (glviewport);

  return PGM_ERROR_OK;
}

static PgmError
pgm_gl_viewport_get_caps_mask (PgmViewport *viewport,
                               gulong *caps_mask)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);

  GST_DEBUG_OBJECT (glviewport, "get_caps_mask");

  GST_OBJECT_LOCK (glviewport);

  /* We don't support shadowing neither stroking */
  *caps_mask = 0x0;

  GST_OBJECT_UNLOCK (glviewport);

  return PGM_ERROR_OK;
}

/* GObject stuff */

G_DEFINE_TYPE (PgmGlViewport, pgm_gl_viewport, PGM_TYPE_VIEWPORT);

static void
pgm_gl_viewport_dispose (GObject *object)
{
  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (object);

  GST_DEBUG_OBJECT (glviewport, "dispose");

  pgm_context_free (glviewport->context);

  GST_CALL_PARENT (G_OBJECT_CLASS, dispose, (object));
}

static void
pgm_gl_viewport_class_init (PgmGlViewportClass *klass)
{
  GObjectClass *gobject_class;
  PgmViewportClass *viewport_class;

  parent_class = g_type_class_peek_parent (klass);

  gobject_class = G_OBJECT_CLASS (klass);
  viewport_class = PGM_VIEWPORT_CLASS (klass);

  /* GObject virtual table */
  gobject_class->dispose = GST_DEBUG_FUNCPTR (pgm_gl_viewport_dispose);

  /* PgmViewport virtual table */
  viewport_class->set_color = GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_color);
  viewport_class->update = GST_DEBUG_FUNCPTR (pgm_gl_viewport_update);
  viewport_class->set_title = GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_title);
  viewport_class->show = GST_DEBUG_FUNCPTR (pgm_gl_viewport_show);
  viewport_class->hide = GST_DEBUG_FUNCPTR (pgm_gl_viewport_hide);
  viewport_class->set_cursor = GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_cursor);
  viewport_class->set_icon = GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_icon);
  viewport_class->set_size = GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_size);
  viewport_class->set_fullscreen =
    GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_fullscreen);
  viewport_class->set_screen_resolution =
    GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_screen_resolution);
  viewport_class->get_screen_resolution =
    GST_DEBUG_FUNCPTR (pgm_gl_viewport_get_screen_resolution);
  viewport_class->set_screen_size_mm =
    GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_screen_size_mm);
  viewport_class->get_screen_size_mm =
    GST_DEBUG_FUNCPTR (pgm_gl_viewport_get_screen_size_mm);
  viewport_class->set_canvas = GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_canvas);
  viewport_class->get_pixel_formats =
    GST_DEBUG_FUNCPTR (pgm_gl_viewport_get_pixel_formats);
  viewport_class->get_caps_mask =
    GST_DEBUG_FUNCPTR (pgm_gl_viewport_get_caps_mask);

  /* PgmGlDrawable changed table */
  klass->changed_func[PGM_DRAWABLE_VISIBILITY] =
    GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_visibility);
  klass->changed_func[PGM_DRAWABLE_SIZE] =
    GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_size);
  klass->changed_func[PGM_DRAWABLE_POSITION] =
    GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_position);
  klass->changed_func[PGM_DRAWABLE_BG_COLOR] =
    GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_bg_color);
  klass->changed_func[PGM_DRAWABLE_FG_COLOR] =
    GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_fg_color);
  klass->changed_func[PGM_DRAWABLE_OPACITY] =
    GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_opacity);
  klass->changed_func[PGM_DRAWABLE_ARC] =
    GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_arc);
  klass->changed_func[PGM_DRAWABLE_STROKE_WIDTH] =
    GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_stroke_width);
  klass->changed_func[PGM_DRAWABLE_STROKE_COLOR] =
    GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_stroke_color);
  klass->changed_func[PGM_DRAWABLE_SHADOW_RADIUS] =
    GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_shadow_radius);
  klass->changed_func[PGM_DRAWABLE_SHADOW_OFFSET] =
    GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_shadow_offset);
  klass->changed_func[PGM_DRAWABLE_SHADOW_COLOR] =
    GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_shadow_color);

  /* PgmGlImage changed table */
  klass->changed_func[PGM_IMAGE_DATA_EMPTY] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_clear);
  klass->changed_func[PGM_IMAGE_DATA_BUFFER] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_from_buffer);
  klass->changed_func[PGM_IMAGE_DATA_GST_BUFFER] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_from_gst_buffer);
  klass->changed_func[PGM_IMAGE_DATA_FD] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_from_fd);
  klass->changed_func[PGM_IMAGE_DATA_IMAGE] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_from_image);
  klass->changed_func[PGM_IMAGE_ALIGNMENT] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_alignment);
  klass->changed_func[PGM_IMAGE_LAYOUT] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_layout);
  klass->changed_func[PGM_IMAGE_INTERP] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_interp);
  klass->changed_func[PGM_IMAGE_ASPECT_RATIO] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_aspect_ratio);

  /* PgmGlText changed table */
  klass->changed_func[PGM_TEXT_LABEL] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_label);
  klass->changed_func[PGM_TEXT_MARKUP] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_markup);
  klass->changed_func[PGM_TEXT_FONT_FAMILY] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_font_family);
  klass->changed_func[PGM_TEXT_HEIGHT] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_font_height);
  klass->changed_func[PGM_TEXT_ELLIPSIZE] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_ellipsize);
  klass->changed_func[PGM_TEXT_JUSTIFY] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_justify);
  klass->changed_func[PGM_TEXT_ALIGNMENT] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_alignment);
  klass->changed_func[PGM_TEXT_WRAP] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_wrap);
  klass->changed_func[PGM_TEXT_GRAVITY] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_gravity);
  klass->changed_func[PGM_TEXT_STRETCH] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_stretch);
  klass->changed_func[PGM_TEXT_STYLE] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_style);
  klass->changed_func[PGM_TEXT_VARIANT] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_variant);
  klass->changed_func[PGM_TEXT_WEIGHT] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_weight);
  klass->changed_func[PGM_TEXT_LINE_SPACING] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_line_spacing);
  klass->changed_func[PGM_TEXT_OUTLINE_COLOR] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_outline_color);
  klass->changed_func[PGM_TEXT_OUTLINE_WIDTH] =
    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_outline_width);
}

static void
pgm_gl_viewport_init (PgmGlViewport *glviewport)
{
  gint screen_width, screen_height;

  GST_DEBUG_OBJECT (glviewport, "init");

  /* Binded canvas */
  glviewport->canvas = NULL;

  /* Empty lists */
  glviewport->far = NULL;
  glviewport->middle = NULL;
  glviewport->near = NULL;

  /* Create the drawable hash using pointers as keys */
  glviewport->drawable_hash = g_hash_table_new (NULL, NULL);

  /* Projected size */
  glviewport->width = 800;
  glviewport->height = 600;
  glviewport->viewport_ratio = 1.333333334f;
  glviewport->canvas_ratio = 1.333333334f;

  /* Create context */
  glviewport->context = pgm_context_new (glviewport);

  /* And update the non square handling parameters */
  pgm_backend_get_screen_size_mm (glviewport->context->backend,
                                  &screen_width, &screen_height);
  update_non_square_parameters (glviewport, screen_width, screen_height);
}

/* Public methods */

PgmViewport *
pgm_gl_viewport_new (void)
{
  PgmGlViewport *glviewport;

  glviewport = g_object_new (PGM_TYPE_GL_VIEWPORT, NULL);
  GST_DEBUG_OBJECT (glviewport, "created new glviewport");

  return PGM_VIEWPORT (glviewport);
}

void
pgm_gl_viewport_update_projection (PgmGlViewport *glviewport,
                                   gint width,
                                   gint height)
{
  GList *list;

  g_return_if_fail (PGM_IS_GL_VIEWPORT (glviewport));

  GST_OBJECT_LOCK (glviewport);

  /* Avoid a whole update if the size is the same */
  if (glviewport->width == width && glviewport->height == height)
    {
      GST_OBJECT_UNLOCK (glviewport);
      return;
    }
  glviewport->width = width;
  glviewport->height = height;

  GST_OBJECT_UNLOCK (glviewport);

  /* Adapt position and size of all the drawables */
  list = glviewport->far;
  while (list)
    {
      pgm_gl_drawable_update_projection (PGM_GL_DRAWABLE (list->data));
      list = list->next;
    }
  list = glviewport->middle;
  while (list)
    {
      pgm_gl_drawable_update_projection (PGM_GL_DRAWABLE (list->data));
      list = list->next;
    }
  list = glviewport->near;
  while (list)
    {
      pgm_gl_drawable_update_projection (PGM_GL_DRAWABLE (list->data));
      list = list->next;
    }
}
